import os, fcntl, errno
import select, compat.select_fix
import logging

try:
    MAXFD = os.sysconf("SC_OPEN_MAX")
except:
    MAXFD = 256

_logger = logging.getLogger("api.util.subproc")


class ExecuteException(Exception):
    def __init__(self, subprocess):
        self.subprocess = subprocess

    def __unicode__(self):
        return u"Subprocess " + unicode(self.subprocess) \
            + self.result() \
            + u"\n== STDOUT ====================\n" + self.subprocess.stdout \
            + u"\n== STDERR ====================\n" + self.subprocess.stderr


class SignalException(ExecuteException):
    def __init__(self, subprocess, signal):
        ExecuteException.__init__(self, subprocess)
        self.signal = signal

    def result(self):
        return u" has been terminated by signal " + unicode(self.signal)


class NonzeroExitException(ExecuteException):
    def __init__(self, subprocess, exitcode):
        ExecuteException.__init__(self, subprocess)
        self.exitcode = exitcode

    def result(self):
        return u" was finished with exit code " + unicode(self.exitcode)

    def __str__(self):
        return self.result().encode("UTF-8")

class BaseSubprocess:
    def __init__(self, prog, args, input, env, keepStderr = False):
        self.__prog = prog
        self.__args = args
        self.__env = env
        self.__input = input
        self.__stdin = self.__stdout = self.__stderr = None
        self.__finished = False
        self.__keepStderr = keepStderr
        self.__forkexec()

    def kill(self, signo):
        """Kills the associated process.

        Kills the associated OS process. wait() should be used to
        collect the process death information.
        """
        if self.__finished:
            _logger.info(u"Skipped killing finished process " + unicode(self.__pid))
        else:
            _logger.info(u"Killing process " + unicode(self.__pid))
            try:
                os.kill(self.__pid, signo)
            except:
                pass

    def __forkexec(self):
        (chstdinr, self.__stdin) = os.pipe()
        (self.__stdout, chstdoutw) = os.pipe()
        if not self.__keepStderr:
            (self.__stderr, chstderrw) = os.pipe()

        self.__pid = os.fork()

        if not self.__pid:
            os.dup2(chstdinr, 0)
            os.dup2(chstdoutw, 1)
            if not self.__keepStderr:
                os.dup2(chstderrw, 2)

            for i in range(3, MAXFD):
                try:
                    os.close(i)
                except:
                    pass

            for i in range(0, len(self.__args)):
                if isinstance(self.__args[i], unicode):
                    self.__args[i] = self.__args[i].encode("UTF-8")

            try:
                os.execve(self.__prog, self.__args, self.__env)
            except OSError, e:
                sys.stderr.write(u"Unable to execute execve: " + unicode(e))
                os.abort()
        else:
            os.close(chstdinr)
            os.close(chstdoutw)
            if not self.__keepStderr:
                os.close(chstderrw)

    def __unicode__(self):
        return u"<subprocess[%d] %s>" % (self.__pid, u" ".join(["'%s'" % x for x in self.__args]))

    def __writeEncodeUtf8(self, fd, val):
        if isinstance(val, unicode):
            val = val.encode("UTF-8")
        try:
            os.write(fd, val)
        except OSError, e:
            if e.errno != errno.EPIPE:
                raise

    def __set_nonblocking(self, fd):
        fl = fcntl.fcntl(fd, fcntl.F_GETFL)
        fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NDELAY)

    def __fetch_data(self, fds):
        out = {}

        p = select.poll()

        for fd in fds:
            out[fd] = []
            self.__set_nonblocking(fd)
            p.register(fd, select.POLLIN)

        fdnum = len(fds)

        while fdnum > 0:
            for (fd, event) in p.poll():
                if event & select.POLLIN:
                    s = os.read(fd, 256)
                    if len(s) == 0:
                        p.unregister(fd)
                        fdnum -= 1
                        continue
                    out[fd].append(s)

                    # clear POLLHUP if POLLIN is here, bc POLLHUP may
                    # arrive before end of data
                    event &= ~(select.POLLIN | select.POLLHUP)

                if event != 0:
                    p.unregister(fd)
                    fdnum -= 1

        for fd in fds:
            out[fd] = "".join(out[fd])
            try:
                out[fd] = out[fd].decode("UTF-8")
            except UnicodeError:
                out[fd] = out[fd].decode("Latin-1")

        return out

    def run(self):
        self.__writeEncodeUtf8(self.__stdin, self.__input)
        os.close(self.__stdin)

        if not self.__keepStderr:
            fds = [self.__stdout, self.__stderr]
        else:
            fds = [self.__stdout]

        res = self.__fetch_data(fds)

        self.stdout = res[self.__stdout]
        if not self.__keepStderr:
            self.stderr = res[self.__stderr]
        else:
            self.stderr = ""

        self.wait()

    def wait(self):
        status = os.waitpid(self.__pid, 0)[1]

        self.__finished = True

        if os.WIFSIGNALED(status):
            raise SignalException(self, os.WTERMSIG(status))
        elif os.WEXITSTATUS(status) != 0:
            raise NonzeroExitException(self, os.WEXITSTATUS(status))


class Subprocess(BaseSubprocess):
    def run(self):
        _logger.info(u"Executing " + unicode(self))
        BaseSubprocess.run(self)

    def wait(self):
        try:
            BaseSubprocess.wait(self)
        except ExecuteException, e:
            _logger.info(unicode(e))
            raise
        else:
            _logger.info(u"Execution of " + unicode(self) + u" finished successfully.")


class CmdLine(object):
    def __init__(self, filename, process_handler = None, args = [], vars = os.environ, stdin = ""):
        self.filename = filename
        self.args = args
        self.vars = vars.copy()
        self.stdin = stdin
        self.process_handler = process_handler
        self.stderr = ""

    def arg(self, arg):
        self.args = self.args + [arg]
        return self

    def var(self, varname, varvalue):
        if isinstance(varvalue, unicode):
            self.vars[varname] = varvalue.encode("UTF-8")
        else:
            self.vars[varname] = varvalue
        return self

    def addstdin(self, stdin_part):
        self.stdin += stdin_part
        return self

    def spawn(self, keepStderr = False):
        proc = Subprocess(self.filename,
                          [self.filename] + self.args,
                          self.stdin,
                          self.vars,
                          keepStderr)
        if self.process_handler:
            self.process_handler.setProcess(proc)

        proc.run()

        if not keepStderr:
            self.stderr = proc.stderr

        if self.process_handler:
            self.process_handler.setProcess(None)

        return proc
