import logging, re, os, errno
import select, compat.select_fix
import plesk_config, gzip_stream, osutil

try: import subprocess
except ImportError: import compat.subprocess as subprocess

_logger = logging.getLogger("api.unpacker")

class Id(object):
    def __init__(self, str_id):
        res = re.search("(.+):gzipped$", str_id)
        if res:
            str_id = res.group(1)
            self.gzipped = True
        else:
            self.gzipped = False

        self.filename = str_id


class Unpacker(object):
    def __call__(self, fd, str_id):
        id = Id(str_id)

        if id.gzipped:
            fd = gzip_stream.GzipReadStream(fd)

        return self.unpack_stream(fd)


def pump(cmd, fd, env = None):
    sub = subprocess.Popen(cmd, stdin = subprocess.PIPE,
                           stdout = subprocess.PIPE, stderr = subprocess.PIPE, env = env)

    read_set = [sub.stdout, sub.stderr]
    write_set = [sub.stdin]

    stdout = ""
    stderr = ""

    stdin_buffer = ""

    while read_set or write_set:
        rlist, wlist, xlist = select.select(read_set, write_set, [])

        # 512-bytes-writes are guaranteed to not block
        # if select returned 'available to write' status
        # on descriptor, so all writes go in the 512-byte chunks.
        #
        # 4096 is the optimal size of fd reading due to internals of MIME-extractor

        if sub.stdout in rlist:
            data = os.read(sub.stdout.fileno(), 1024)
            if data == "":
                sub.stdout.close()
                read_set.remove(sub.stdout)
            stdout += data
            continue

        if sub.stderr in rlist:
            data = os.read(sub.stderr.fileno(), 1024)
            if data == "":
                sub.stderr.close()
                read_set.remove(sub.stderr)
            stderr += data
            continue

        if sub.stdin in wlist:
            if len(stdin_buffer) < 512:
                if fd:
                    data = fd.read(4096)
                    if data == "":
                        fd = None
                    stdin_buffer += data

            if not len(stdin_buffer):
                sub.stdin.close()
                write_set.remove(sub.stdin)
            else:
                try:
                    written = os.write(sub.stdin.fileno(), stdin_buffer[:512])
                except OSError, e:
                    if e.errno == errno.EPIPE:
                        sub.stdin.close()
                        write_set.remove(sub.stdin)
                    else:
                        raise
                else:
                    stdin_buffer = stdin_buffer[written:]


    retvalue = sub.wait()

    return (retvalue, stdout, stderr)

class DbUnpacker(Unpacker):
    def __init__(self, name, type, user, password, host, port):
        self.name = name
        self.type = type
        self.user = user
        self.password = password
        self.host = host
        self.port = port

    def unpack_stream(self, fd):
        if self.type == "mysql":
            env = os.environ.copy()
            env["MYSQL_PWD"] = self.password

            cmd = [os.path.join(plesk_config.get("MYSQL_BIN_D"), "mysql")]
            cmd.append("--no-defaults")
            cmd.append("-u" + self.user)

            if self.host != "":
                cmd.append("-h")
                cmd.append(self.host)
            if self.port != "":
                cmd.append("-P")
                cmd.append(self.port)
            cmd.append(self.name)

            retvalue, stdout, stderr = pump(cmd, fd, env)

            if retvalue != 0:
                raise ValueError(retvalue, stdout, stderr)
        else:
            env = os.environ.copy()
            env["PGUSER"] = self.user
            env["PGPASSWORD"] = self.password

            # option --use-set-session-authorization for old version of postgresql dump, bug 90676
            cmd = [os.path.join(plesk_config.get("PGSQL_BIN_D"), "pg_restore"), "-d", self.name,
                   "--use-set-session-authorization"]
            if self.host != "" and self.host != "localhost":
                cmd.append("-h")
                cmd.append(self.host)
            if self.port != "":
                cmd.append("-p")
                cmd.append(self.port)

            try:
                retvalue, stdout, stderr = pump(cmd, fd, env)
            except OSError, e:
                if e.errno == errno.ENOENT:
                    raise RuntimeError("No postgresql command-line utility installed")
                raise

            if retvalue != 0:
                raise ValueError(retvalue, stdout, stderr)

        return []

class TarUnpacker(Unpacker):
    def __init__(self, destdir):
        self.destdir = destdir

    def unpack_stream(self, fd):

        retvalue, stdout, stderr = pump([plesk_config.get("TAR_BIN"), "-xvf", "-",
                                         "-C", self.destdir], fd)

        if retvalue != 0:
            raise ValueError(retvalue, stdout, stderr)

        # TAR_BIN might be GNU or BSD tar.
        return parseGnuTarOutput(stdout) + parseBsdTarOutput(stderr)

# GNU tar: stdout, line per file
def parseGnuTarOutput(s):
    return s.splitlines()

# BSD tar: stderr, file prepended with 'x ' string
def parseBsdTarOutput(s):
    return [line[2:] for line in s.splitlines() if line.startswith('x ')]

class PartUnpacker(Unpacker):
    def __init__(self, destfile):
        self.destfile = destfile

    def unpack_stream(self, fh):
        osutil.saveToFile(self.destfile, fh)
        return []
